import math
import torch
import torch.nn as nn
import torch.nn.functional as F


__all__ = ['MultiHeadAttention', 'ScaledDotProductAttention']


class ScaledDotProductAttention(nn.Module):

    def forward(self, query, key, value, mask=None):
        dk = query.size()[-1]
        scores = query.matmul(key.transpose(-2, -1)) / math.sqrt(dk)
        if mask is not None:
            scores = scores.masked_fill(mask == 0, -1e9)
        attention = F.softmax(scores, dim=-1)
        return attention.matmul(value)


class MultiHeadAttention(nn.Module):

    def __init__(self,
                 in_features,
                 head_num,
                 bias=True,
                 activation=F.relu):
        """Multi-head attention.

        :param in_features: Size of each input sample.
        :param head_num: Number of heads.
        :param bias: Whether to use the bias term.
        :param activation: The activation after each linear transformation.
        """
        super(MultiHeadAttention, self).__init__()
        if in_features % head_num != 0:
            raise ValueError('`in_features`({}) should be divisible by `head_num`({})'.format(in_features, head_num))
        self.in_features = in_features
        self.head_num = head_num
        self.activation = activation
        self.bias = bias
        self.linear_q = nn.Linear(in_features, in_features, bias)
        self.linear_k = nn.Linear(in_features, in_features, bias)
        self.linear_v = nn.Linear(in_features, in_features, bias)
        self.linear_o = nn.Linear(in_features, in_features, bias)

    def forward(self, q, k, v, mask=None):
        q, k, v = self.linear_q(q), self.linear_k(k), self.linear_v(v)
        if self.activation is not None:
            q = self.activation(q)
            k = self.activation(k)
            v = self.activation(v)

        q = self._reshape_to_batches(q)
        k = self._reshape_to_batches(k)
        v = self._reshape_to_batches(v)
        if mask is not None:
            mask = mask.repeat(self.head_num, 1, 1)
        y = ScaledDotProductAttention()(q, k, v, mask)
        y = self._reshape_from_batches(y)

        y = self.linear_o(y)
        if self.activation is not None:
            y = self.activation(y)
        return y

    @staticmethod
    def gen_history_mask(x):
        """Generate the mask that only uses history data.

        :param x: Input tensor.
        :return: The mask.
        """
        batch_size, seq_len, _ = x.size()
        return torch.tril(torch.ones(seq_len, seq_len)).view(1, seq_len, seq_len).repeat(batch_size, 1, 1)

    def _reshape_to_batches(self, x):
        batch_size, seq_len, in_feature = x.size()
        sub_dim = in_feature // self.head_num
        return x.reshape(batch_size, seq_len, self.head_num, sub_dim)\
                .permute(0, 2, 1, 3)\
                .reshape(batch_size * self.head_num, seq_len, sub_dim)

    def _reshape_from_batches(self, x):
        batch_size, seq_len, in_feature = x.size()
        batch_size //= self.head_num
        out_dim = in_feature * self.head_num
        return x.reshape(batch_size, self.head_num, seq_len, in_feature)\
                .permute(0, 2, 1, 3)\
                .reshape(batch_size, seq_len, out_dim)

    def extra_repr(self):
        return 'in_features={}, head_num={}, bias={}, activation={}'.format(
            self.in_features, self.head_num, self.bias, self.activation,
        )



import os
import tempfile
import random
from unittest import TestCase
import torch
import keras
import numpy as np
from keras_multi_head import MultiHeadAttention as KerasMultiHeadAttention
from torch_multi_head_attention import MultiHeadAttention


class TestMultiHeadAttention(TestCase):

    def test_divisible(self):
        with self.assertRaises(ValueError):
            MultiHeadAttention(in_features=73, head_num=5)

    @staticmethod
    def get_torch_layer_with_weights(feature_dim, head_num, weights, bias):
        layer = MultiHeadAttention(feature_dim, head_num)
        layer.linear_q.weight = torch.nn.Parameter(
            torch.from_numpy(weights[:, :feature_dim]).transpose(1, 0)
        )
        layer.linear_q.bias = torch.nn.Parameter(
            torch.from_numpy(bias[:feature_dim])
        )
        layer.linear_k.weight = torch.nn.Parameter(
            torch.from_numpy(weights[:, feature_dim:feature_dim * 2]).transpose(1, 0)
        )
        layer.linear_k.bias = torch.nn.Parameter(
            torch.from_numpy(bias[feature_dim:feature_dim * 2])
        )
        layer.linear_v.weight = torch.nn.Parameter(
            torch.from_numpy(weights[:, feature_dim * 2:feature_dim * 3]).transpose(1, 0)
        )
        layer.linear_v.bias = torch.nn.Parameter(
            torch.from_numpy(bias[feature_dim * 2:feature_dim * 3])
        )
        layer.linear_o.weight = torch.nn.Parameter(
            torch.from_numpy(weights[:, -feature_dim:]).transpose(1, 0)
        )
        layer.linear_o.bias = torch.nn.Parameter(
            torch.from_numpy(bias[-feature_dim:])
        )
        return layer

    @staticmethod
    def get_keras_layer_weight_weights(seq_len, feature_dim, head_num, weights, bias, history_only=False):
        input_layer = keras.layers.Input(shape=(seq_len, feature_dim), name='Input')
        attention_layer = KerasMultiHeadAttention(
            head_num,
            history_only=history_only,
            weights=[
                weights[:, :feature_dim],
                bias[:feature_dim],
                weights[:, feature_dim:feature_dim * 2],
                bias[feature_dim:feature_dim * 2],
                weights[:, feature_dim * 2:feature_dim * 3],
                bias[feature_dim * 2:feature_dim * 3],
                weights[:, -feature_dim:],
                bias[-feature_dim:],
            ],
        )(input_layer)
        model = keras.models.Model(input_layer, attention_layer)
        model.compile(optimizer='adam', loss='mse')
        return model

    def test_same_output_without_mask(self):
        batch_size, seq_len, feature_dim, head_num = 7, 12, 16, 4
        weights = np.random.standard_normal((feature_dim, feature_dim * 4))
        bias = np.random.standard_normal((feature_dim * 4,))
        torch_net = self.get_torch_layer_with_weights(feature_dim, head_num, weights, bias)
        keras_net = self.get_keras_layer_weight_weights(seq_len, feature_dim, head_num, weights, bias)
        print(torch_net)
        allclose_count = 0
        for _ in range(100):
            x = np.random.standard_normal((batch_size, seq_len, feature_dim))
            y = keras_net.predict(x)
            x = torch.from_numpy(x)
            y_hat = torch_net(x, x, x)
            if np.allclose(y, y_hat.detach().numpy(), rtol=0.0, atol=1e-4):
                allclose_count += 1
        self.assertGreaterEqual(allclose_count, 98)

    def test_same_output_history_only(self):
        batch_size, seq_len, feature_dim, head_num = 7, 12, 16, 4
        weights = np.random.standard_normal((feature_dim, feature_dim * 4))
        bias = np.random.standard_normal((feature_dim * 4,))
        torch_net = self.get_torch_layer_with_weights(feature_dim, head_num, weights, bias)
        keras_net = self.get_keras_layer_weight_weights(seq_len, feature_dim, head_num, weights, bias, True)
        allclose_count = 0
        for _ in range(100):
            x = np.random.standard_normal((batch_size, seq_len, feature_dim))
            y = keras_net.predict(x)
            x = torch.from_numpy(x)
            y_hat = torch_net(x, x, x, MultiHeadAttention.gen_history_mask(x))
            if np.allclose(y, y_hat.detach().numpy(), rtol=0.0, atol=1e-4):
                allclose_count += 1
        self.assertGreaterEqual(allclose_count, 98)



from unittest import TestCase
import torch
from torch_multi_head_attention import ScaledDotProductAttention, MultiHeadAttention


class TestScaledDotProductAttention(TestCase):

    def test_sample(self):
        x = torch.Tensor([[
            [0.2, 0.3, 0.4, 0.6, 0.5],
            [0.4, 0.7, 0.2, 0.6, 0.9],
            [0.3, 0.5, 0.8, 0.9, 0.1],
            [0.2, 0.3, 0.4, 0.6, 0.5],
            [0.1, 0.2, 0.3, 0.4, 0.5],
        ]])
        mask = torch.Tensor([[1, 1, 1, 1, 0]])
        y = ScaledDotProductAttention()(x, x, x, mask)[0]
        self.assertTrue(y[0].allclose(y[3]), y)
        self.assertTrue(y[2].allclose(torch.Tensor([0.27883747, 0.45767492, 0.47448885, 0.69199574, 0.47368336])), y[2])

    def test_history_only(self):
        x = torch.Tensor([[
            [0.2, 0.3, 0.4, 0.6, 0.5],
            [0.4, 0.7, 0.2, 0.6, 0.9],
            [0.3, 0.5, 0.8, 0.9, 0.1],
            [0.2, 0.3, 0.4, 0.6, 0.5],
            [0.1, 0.2, 0.3, 0.4, 0.5],
        ]])
        mask = MultiHeadAttention.gen_history_mask(x)
        y = ScaledDotProductAttention()(x, x, x, mask)[0]
        self.assertFalse(y[0].allclose(y[3]), y)
        self.assertTrue(y[0].allclose(x[0, 0]), y[0])
